/*
 * @(#)AbacusE.c
 *
 * Copyright 2017  David A. Bagley, bagleyd AT verizon.net
 *
 * Abacus examples
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Method to handle the examples for Abacus */

/**
 * Much help from: Dodji Seketeli
 * http://www.xmlsoft.org/examples/tree1.c
 */

#include "AbacusP.h"

/* This can probably be fixed to use libxml2 Windows port */
/* https://www.zlatkovic.com/libxml.en.html */
#ifdef HAVE_LIBXML2
#include <libxml/parser.h>
#include <libxml/tree.h>
#endif

#ifdef LIBXML_TREE_ENABLED

static Boolean DEBUG = False;
/*static Boolean DEBUG = True;*/

unsigned int codeMaxCount = 0; /*4*/
unsigned int moveMaxCount = 0; /*41*/
unsigned int lessonMaxCount = 0; /*7*/
unsigned int chapterMaxCount = 0; /*7*/
unsigned int editionMaxCount = 0; /*4*/
unsigned int bookMaxCount = 0; /*3*/

xmlDoc *doc = NULL;

/* From AndiDog
 https://stackoverflow.com/questions/2450704/writing-string-trim-in-c */
static void
trim(char *str)
{
	char *ptr = str;
	while (*ptr == ' ' || *ptr == '\t' || *ptr == '\r' || *ptr == '\n')
		++ptr;

	char *end = ptr;
	while (*end)
		++end;

	if (end > ptr) {
		for (--end; end >= ptr && (*end == ' ' || *end == '\t' || *end == '\r' || *end == '\n'); --end);
	}
	++end; /* was loosing last character */
	memmove(str, ptr, (size_t) (end - ptr));
	str[end - ptr] = 0;
} 

/* Alexey Frunze
https://stackoverflow.com/questions/8512958/is-there-a-windows-variant-of-strsep */
static char *
str_sep(char **stringp, const char *delim)
{
	char* start = *stringp;
	char* p;

	p = (start != NULL) ? strpbrk(start, delim) : NULL;

	if (p == NULL) {
		*stringp = NULL;
	} else {
		*p = '\0';
		*stringp = p + 1;
	}
	return start;
}

/* From Costya Perepelitsa
 https://www.quora.com/How-do-you-write-a-C-program-to-split-a-string-by-a-delimiter */
static char **
strsplit(const char* str, const char* delim, size_t* numtokens) {
	char *s = strdup(str);
	size_t tokens_alloc = 1;
	size_t tokens_used = 0;
	char **tokens = calloc(tokens_alloc, sizeof(char*));
	char *token, *rest = s;
	while ((token = str_sep(&rest, delim)) != NULL) {
		if (tokens_used == tokens_alloc) {
			tokens_alloc *= 2;
			tokens = realloc(tokens, tokens_alloc * sizeof(char*));
		}
		tokens[tokens_used++] = strdup(token);
	}
	if (tokens_used == 0) {
		free(tokens);
		tokens = NULL;
	} else {
		tokens = realloc(tokens, tokens_used * sizeof(char*));
	}
	*numtokens = tokens_used;
	free(s);
	return tokens;
}

static void
codeCount(xmlNode *topNode)
{
	xmlNode *node = NULL;
	int term = 0;

	for (node = topNode->children; node; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (term == 1) {
				xmlChar *value = xmlNodeListGetString(doc,
					node->xmlChildrenNode, 1);
				char * tmp = (char *) malloc(sizeof(char) *
					strlen((char *) value) + 1);
				(void) strcpy(tmp, (char *) value);
				xmlFree(value);
				trim(tmp);
				char *tmp2 = tmp;
				int i;
				for (i = 0; tmp[i]; (tmp[i] == '\n') ? i++ : *tmp++);
				free(tmp2);
				if (codeMaxCount < (unsigned int) i) {
					codeMaxCount = (unsigned int) i;
				}
			}
			term++;
		}
	}
}

static void
moveCount(xmlNode *topNode)
{
	xmlNode *node = NULL;
	unsigned int move = 0;

	for (node = topNode->children; node; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			codeCount(node);
			move++;
		}
	}
	if (moveMaxCount < move) {
		moveMaxCount = move;
	}
}

static void
lessonCount(xmlNode *topNode)
{
	xmlNode *node = NULL;
	unsigned int lesson = 0;

	for (node = topNode->children; node; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			moveCount(node);
			lesson++;
		}
	}
	if (lessonMaxCount < lesson) {
		lessonMaxCount = lesson;
	}
}

static void
chapterCount(xmlNode *topNode)
{
	xmlNode *node = NULL;
	unsigned int chapter = 0;

	for (node = topNode->children; node; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			lessonCount(node);
			chapter++;
		}
	}
	if (chapterMaxCount < chapter) {
		chapterMaxCount = chapter;
	}
}

static void
editionCount(xmlNode *topNode)
{
	xmlNode *node = NULL;
	unsigned int edition = 0;

	for (node = topNode->children; node; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			chapterCount(node);
			edition++;
		}
	}
	if (editionMaxCount < edition) {
		editionMaxCount = edition;
	}
}

static void
bookCount(xmlNode *topNode)
{
	xmlNode *node = NULL;
	unsigned int book = 0;

	for (node = topNode->children; node; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			editionCount(node);
			book++;
		}
	}
	if (bookMaxCount < book) {
		bookMaxCount = book;
	}
}

static struct moveType
readMove(xmlNode *topNode)
{
	struct moveType * moveAbacus = (struct moveType *)
		malloc(sizeof(struct moveType));
	xmlNode *node = NULL;
	const int moveItems = 2;
	char *tmp[moveItems];
	int term = 0;

	moveAbacus->code = (char *) malloc(sizeof(char));
	moveAbacus->lineText = (char **) malloc(sizeof(char *) * (codeMaxCount + 1));
	if (DEBUG) {
		(void) printf("\t\t\t\t\t%s\n", topNode->name);
	}
	for (node = topNode->children; node && term < moveItems; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			xmlChar *value = xmlNodeListGetString(doc,
				node->xmlChildrenNode, 1);
			tmp[term] = (char *) malloc(sizeof(char) *
				strlen((char *) value) + 1);
			(void) strcpy(tmp[term], (char *) value);
			xmlFree(value);
			if (DEBUG) {
				(void) printf("\t\t\t\t\t%s\n", node->name);
				(void) printf("value: %s\n", tmp[term]);
			}
			if (term == 0) {
				moveAbacus->code = tmp[0];
			} else if (term == 1) {
				unsigned int i;
				size_t lines;

				trim(tmp[1]);
				char **text = strsplit(tmp[1], "\n", &lines);
				for (i = 0; i < lines && i < codeMaxCount; i++) {
					trim(text[i]);
					moveAbacus->lineText[i] = (char *) malloc(sizeof(char) *
						strlen((char *) text[i]) + 1);
					(void) strcpy(moveAbacus->lineText[i], text[i]);
				}
				free(tmp[1]);
				moveAbacus->lines = lines;
				moveAbacus->lineText[lines] = NULL;
			}
			term++;
		}
	}
	return *moveAbacus;
}

static struct lessonType
readLesson(xmlNode *topNode)
{
	struct lessonType * lessonAbacus = (struct lessonType *)
		malloc(sizeof(struct lessonType));
	xmlNode *node = NULL;
	unsigned int move = 0;
	xmlChar *name = xmlGetProp(topNode, (const xmlChar *) "name");

	if (DEBUG) {
		(void) printf("\t\t\t\t%s\n", topNode->name);
		(void) printf("\t\t\t\t    name: %s\n", name);
	}
	lessonAbacus->name = (char *) malloc(sizeof(char) *
		strlen((char *) name) + 1);
	(void) strcpy(lessonAbacus->name, (char *) name);
	xmlFree(name);
	lessonAbacus->move = (struct moveType *)
		malloc(sizeof(struct moveType) * (moveMaxCount + 1));
	for (node = topNode->children; node && move <= moveMaxCount; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			lessonAbacus->move[move] = readMove(node);
			move++;
		}
	}
	lessonAbacus->moves = move;
	lessonAbacus->move[move].code = NULL;
	lessonAbacus->move[move].lines = 0;
	lessonAbacus->move[move].lineText = NULL;
	return *lessonAbacus;
}

static struct chapterType
readChapter(xmlNode *topNode)
{
	struct chapterType * chapterAbacus = (struct chapterType *)
		malloc(sizeof(struct chapterType));
	xmlNode *node = NULL;
	unsigned int lesson = 0;
	xmlChar *name = xmlGetProp(topNode, (const xmlChar *) "name");

	if (DEBUG) {
		(void) printf("\t\t\t%s\n", topNode->name);
		(void) printf("\t\t\t    name: %s\n", name);
	}
	chapterAbacus->name = (char *) malloc(sizeof(char) *
		strlen((char *) name) + 1);
	(void) strcpy(chapterAbacus->name, (char *) name);
	xmlFree(name);
	chapterAbacus->lesson = (struct lessonType *)
		malloc(sizeof(struct lessonType) * (lessonMaxCount + 1));
	for (node = topNode->children; node && lesson < lessonMaxCount; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (DEBUG) {
				(void) printf("\t\t\t\t%d\n", lesson);
			}
			chapterAbacus->lesson[lesson] = readLesson(node);
			lesson++;
		}
	}
	chapterAbacus->lessons = lesson;
	chapterAbacus->lesson[lesson].name = NULL;
	chapterAbacus->lesson[lesson].moves = 0;
	chapterAbacus->lesson[lesson].move = NULL;
	return *chapterAbacus;
}

static struct editionType
readEdition(xmlNode *topNode)
{
	struct editionType * editionAbacus = (struct editionType *)
		malloc(sizeof(struct editionType));
	xmlNode *node = NULL;
	unsigned int chapter = 0;
	xmlChar *version = xmlGetProp(topNode, (const xmlChar *) "version");

	if (DEBUG) {
		(void) printf("\t\t%s\n", topNode->name);
		(void) printf("\t\t    version: %s\n", version);
	}
	editionAbacus->version = (char *) malloc(sizeof(char) *
		strlen((char *) version) + 1);
	(void) strcpy(editionAbacus->version, (char *) version);
	xmlFree(version);
	editionAbacus->chapter = (struct chapterType *)
		malloc(sizeof(struct chapterType) * (chapterMaxCount + 1));
	for (node = topNode->children; node && chapter < chapterMaxCount; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (DEBUG) {
				(void) printf("\t\t\t%d\n", chapter);
			}
			editionAbacus->chapter[chapter] = readChapter(node);
			chapter++;
		}
	}
	editionAbacus->chapters = chapter;
	editionAbacus->chapter[chapter].name = NULL;
	editionAbacus->chapter[chapter].lessons = 0;
	editionAbacus->chapter[chapter].lesson = NULL;
	return *editionAbacus;
}

static struct bookType
readBook(xmlNode *topNode)
{
	struct bookType * bookAbacus = (struct bookType *)
		malloc(sizeof(struct bookType));
	xmlNode *node = NULL;
	unsigned int edition = 0;
	xmlChar *name = xmlGetProp(topNode, (const xmlChar *) "name");
	xmlChar *author = xmlGetProp(topNode, (const xmlChar *) "author");

	if (DEBUG) {
		(void) printf("\t%s\n", topNode->name);
		(void) printf("\t    name: %s\n", name);
		(void) printf("\t    author: %s\n", author);
	}
	bookAbacus->name = (char *) malloc(sizeof(char) *
		strlen((char *) name) + 1);
	(void) strcpy(bookAbacus->name, (char *) name);
	bookAbacus->author = (char *) malloc(sizeof(char) *
		strlen((char *) author) + 1);
	(void) strcpy(bookAbacus->author, (char *) author);
	xmlFree(name);
	xmlFree(author);
	bookAbacus->edition = (struct editionType *)
		malloc(sizeof(struct editionType) * (editionMaxCount + 1));
	for (node = topNode->children; node && edition < editionMaxCount; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (DEBUG) {
				(void) printf("\t\t%d\n", edition);
			}
			bookAbacus->edition[edition] = readEdition(node);
			edition++;
		}
	}
	bookAbacus->editions = edition;
	bookAbacus->edition[edition].version = NULL;
	bookAbacus->edition[edition].chapters = 0;
	bookAbacus->edition[edition].chapter = NULL;
	return *bookAbacus;
}

static struct libraryType
readAbacus(xmlNode *topNode)
{
	struct libraryType * libraryAbacus = (struct libraryType *)
		malloc(sizeof(struct libraryType));
	xmlNode *node = NULL;
	unsigned int book = 0;

	libraryAbacus->book = (struct bookType *)
		malloc(sizeof(struct bookType) * (bookMaxCount + 1));
	for (node = topNode->children; node && book < bookMaxCount; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (DEBUG) {
				(void) printf("\t%d\n", book);
			}
			libraryAbacus->book[book] = readBook(node);
			book++;
		}
	}
	libraryAbacus->books = book;
	libraryAbacus->book[book].name = NULL;
	libraryAbacus->book[book].author = NULL;
	libraryAbacus->book[book].editions = 0;
	libraryAbacus->book[book].edition = NULL;
	return *libraryAbacus;
}

static void
setLibrary(xmlDoc *document) {
	/* Get the root element node */
	xmlNode *topNode = xmlDocGetRootElement(document);
	xmlNode *node = NULL;

	for (node = topNode; node; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			bookCount(node);
		}
	}
	if (DEBUG) {
		(void) printf("%d %d %d %d %d %d\n",
			codeMaxCount, moveMaxCount, lessonMaxCount,
			chapterMaxCount, editionMaxCount, bookMaxCount);
	}
	for (node = topNode; node; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (DEBUG) {
				printf("%s\n", node->name);
			}
			abacusLibrary = readAbacus(node);
		}
	}

	/* Free the document */
	xmlFreeDoc(document);

	/*
	 * Free the global variables that may
	 * have been allocated by the parser.
	 */
	xmlCleanupParser();
}

static xmlDoc *
getDoc(const char * fileName) {
	doc = xmlReadFile(fileName, NULL, 0);
	if (doc == NULL) {
		(void) fprintf(stderr, "Could not parse file %s\n", fileName);
	}
	return doc;
}

void
readParseFile(const char * fileName) {
	/* treating as a singleton */
	if (abacusLibrary.book != NULL)
		return;
	doc = getDoc(fileName);
	if (doc != NULL)
		setLibrary(doc);
}

/**
 * Simple example to parse a file called "file.xml", 
 * walk down the DOM, and print the name of the 
 * xml elements nodes.
 */
#if 0
char * xmlFile = XML_FILE;
int
emain(int argc, char **argv)
{
	if (argc >= 2) {
		xmlFile = argv[1];
	}

	/*
	 * This initialize the library and check potential ABI mismatches
	 * between the version it was compiled for and the actual shared
	 * library used.
	 */
	LIBXML_TEST_VERSION
	readParseFile(XML_FILE);
	return 0;
}
#endif
#else
void
readParseFile(const char * fileName) {
#ifndef WINVER
	(void) fprintf(stderr, "XML tree support not compiled in\n");
#endif
}
#if 0
int emain(void)
{
	(void) fprintf(stderr, "XML tree support not compiled in\n");
	exit(1);
}
#endif
#endif
